import argparse
from typing import Any
from ._send import Sender
from ._receive import Receiver
from ._exceptions import GgArgumentsError
from . import __version__


class GgHelpFormatter(argparse.RawTextHelpFormatter):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def format_help(self) -> str:
        help_msg = self._root_section.format_help()
        if help_msg:
            help_msg = self._long_break_matcher.sub('\n\n', help_msg)
            help_msg = help_msg.strip('\n') + '\n'
        return help_msg


def is_postive_int(val: int) -> int:
    # try:
    val_int = int(val)
    if val_int > 0:
        return val_int
    raise argparse.ArgumentTypeError("number of pieces must be a positive integer.")


def _main() -> None:

    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(prog="gg-transfer",
                                     # formatter_class=GgHelpFormatter,
                                     description="Command line utility to send/receive "
                                                 "files/strings via ggwave library (FSK).")
    parser.add_argument(
        "-V", "--version",
        help="print version number.",
        action="version", version=f"%(prog)s {__version__}")

    subparsers = parser.add_subparsers(required=True, title="commands",
                                       help="send or receive data.")

    # noinspection PyTypeChecker
    sender = subparsers.add_parser(
        "send", help="modulate data into audio signal.", formatter_class=GgHelpFormatter,
        description="Command line utility to send/receive files/strings via ggwave library (FSK).")
    sender.add_argument(
        "-i", "--input", help="input file (use '-' for stdin).", metavar="<inputfile>")
    sender.add_argument(
        "-p", "--protocol", help="protocol, 0 to 8 (defaults to %(default)s)\n"
                                 "0 = Normal (11,17 Bytes/s - 1875 Hz to 6375 Hz)\n"
                                 "1 = Fast (16,76 Bytes/s - 1875 Hz to 6375 Hz)\n"
                                 "2 = Fastest (33,52 Bytes/s 1875 Hz to 6375 Hz)\n"
                                 "3 = [U] Normal (11,17 Bytes/s - 15000 Hz to 19500 Hz)\n"
                                 "4 = [U] Fast (16,76 Bytes/s - 15000 Hz to 19500 Hz)\n"
                                 "5 = [U] Fastest (33,52 Bytes/s - 15000 Hz to 19500 Hz)\n"
                                 "6 = [DT] Normal (3,72 Bytes/s - 1125 Hz to 2625 Hz)\n"
                                 "7 = [DT] Fast (5,59 Bytes/s - 1125 Hz to 2625 Hz)\n"
                                 "8 = [DT] Fastest (11,17 Bytes/s - 1125 Hz to 2625 Hz)",
        default=0,
        type=int,
        choices=range(0, 9)
    )
    sender.add_argument(
        "-f", "--file-transfer",
        help="encode data in Base64 and use file transfer mode.",
        action="store_true", default=False)
    sender.set_defaults(command="send")

    # noinspection PyTypeChecker
    receiver = subparsers.add_parser(
        "receive", help="demodulate data from audio signal.", formatter_class=GgHelpFormatter,
        description="Command line utility to send/receive files/strings via ggwave library (FSK).")
    receiver.add_argument(
        "-o", "--output", help="output file (use '-' for stdout).", metavar="<outputfile>")
    receiver.add_argument(
        "-f", "--file-transfer",
        help="decode data from Base64 and use file transfer mode.",
        action="store_true", default=False)
    receiver.add_argument(
        "-w", "--overwrite",
        help="overwrite output file if it exists.",
        action="store_true", default=False)
    receiver.add_argument(
        "-n", "--tot-pieces",
        help="receive this number of pieces and exit. Minimum is 1, default no limit.",
        default=-1, type=is_postive_int)

    receiver.set_defaults(command="receive")

    args: argparse.Namespace = parser.parse_args()

    if args.command == "send":
        Sender(args).send()
    elif args.command == "receive":
        Receiver(args).receive(getdata=False)
    else:
        raise GgArgumentsError("No such command.")


if __name__ == '__main__':
    _main()



import tempfile
import unittest
import wave
from pathlib import Path

import pyaudio


# noinspection PyPep8Naming
class PyAudio(unittest.TestCase):

    def test_record(self) -> None:
        CHUNK = 1024
        FORMAT = pyaudio.paInt32
        CHANNELS = 1
        RATE = 48000
        RECORD_SECONDS = 20
        out_file = Path(tempfile.gettempdir()).absolute().joinpath('test.wav')
        wf: wave.Wave_write
        with wave.open(str(out_file), 'w') as wf:

            p = pyaudio.PyAudio()
            wf.setnchannels(CHANNELS)
            wf.setsampwidth(p.get_sample_size(FORMAT))
            wf.setframerate(RATE)

            stream = p.open(format=FORMAT, channels=CHANNELS, rate=RATE, input=True)
            print(f'Recording {RECORD_SECONDS} secs into {out_file}...')
            for _ in range(0, RATE // CHUNK * RECORD_SECONDS):
                wf.writeframes(stream.read(CHUNK))
            print('Done')
            stream.stop_stream()
            stream.close()
            p.terminate()
        self.assertEqual(True, True)  # add assertion here


if __name__ == '__main__':
    unittest.main()



class GgTransferError(Exception):

    msg: str
    _PREFIX: str

    def __init__(self, message: str):
        self._PREFIX = "*** ERROR: "
        self.msg = self._PREFIX + message

    def __str__(self) -> str:
        return self.msg


class GgIOError(GgTransferError):

    _PREFIX: str = "I/O - "

    def __init__(self, message: str):
        super().__init__(self._PREFIX + message)


class GgUnicodeError(GgTransferError):

    _PREFIX: str = "Data type mismatch - "

    def __init__(self, message: str):
        super().__init__(self._PREFIX + message)


class GgArgumentsError(GgTransferError):

    _PREFIX: str = "Invalid arguments - "

    def __init__(self, message: str):
        super().__init__(self._PREFIX + message)


class GgChecksumError(GgTransferError):

    _PREFIX: str = "Checksum error - "

    def __init__(self, message: str):
        super().__init__(self._PREFIX + message)



import argparse
import base64
import binascii
import sys
import time
import pyaudio
import ggwave
from pathlib import Path
from typing import List, Optional, Tuple
from ._exceptions import GgIOError, GgUnicodeError, GgArgumentsError


def _get_array(data: str, crc: bool = False) -> Tuple[List[str], int]:
    siz = 132 if crc else 140
    ar = [data[i:i + siz] for i in range(0, len(data), siz)]
    ln = len(ar)
    crc_blocks: List[str] = []
    if crc:
        for i in range(0, ln):
            crc32: int = binascii.crc32(ar[i].encode(encoding="utf-8"))
            fixed_length_hex: str = f'{crc32:08x}'
            crc_blocks.append(fixed_length_hex)
        for i in range(1, ln):
            ar[i] = crc_blocks[i-1] + ar[i]
        ar[0] = crc_blocks[ln-1] + ar[0]
    return ar, ln


class Sender:

    def __init__(self, args: Optional[argparse.Namespace] = None, inputfile: Optional[str] = None,
                 protocol: int = 0, file_transfer: bool = False):

        if args is not None and isinstance(args, argparse.Namespace):
            self.protocol = args.protocol
            self.file_transfer_mode = args.file_transfer
            self.crc = self.file_transfer_mode
            self.input = args.input
            self._script = True
        elif args is None:
            self.protocol = protocol
            self.file_transfer_mode = file_transfer
            self.crc = self.file_transfer_mode
            self.input = inputfile
            self._script = False
        else:
            raise GgArgumentsError("Wrong set of arguments.")

        self._sample_rate = 48000

    def send(self, msg: Optional[str] = None) -> None:
        p: Optional[pyaudio.PyAudio] = None
        stream: Optional[pyaudio.Stream] = None

        try:
            p = pyaudio.PyAudio()

            # 0 = Normal
            # 1 = Fast
            # 2 = Fastest
            # 3 = [U] Normal
            # 4 = [U] Fast
            # 5 = [U] Fastest
            # 6 = [DT] Normal
            # 7 = [DT] Fast
            # 8 = [DT] Fastest

            stream = p.open(format=pyaudio.paFloat32, channels=1, rate=self._sample_rate, output=True,
                            frames_per_buffer=4096)
            if self.input is not None and self.input != "-" and msg is None:
                file_path = Path(self.input)
                if not file_path.is_file():
                    raise GgIOError(f"File {file_path.absolute()} does not exist.")
                s = file_path.stat()
                size = s.st_size
                name = file_path.name
                with open(file_path, "rb", buffering=0) as f:
                    if self.file_transfer_mode:
                        base_binary = f.read()
                        base = base64.urlsafe_b64encode(base_binary).decode("utf-8")
                        crc32_c: int = binascii.crc32(base_binary)
                        fixed_length_hex: str = f'{crc32_c:08x}'
                        ar, ln = _get_array(base, crc=self.crc)
                        header = ('{0}"pieces": {1}, "filename": "{2}", "size": {3},'
                                  ' "crc": "{4}"{5}').format(
                            "{", str(ln), name, str(size), fixed_length_hex, "}"
                        )
                        stream.write(b'0' * 4 * self._sample_rate * 1)
                        print("Sending header, length:", len(header), flush=True, file=sys.stderr)
                        print("Pieces:", ln, flush=True, file=sys.stderr)
                        waveform = ggwave.encode(header, protocolId=self.protocol, volume=60)
                        stream.write(waveform, len(waveform) // 4)
                    else:
                        # print("Only the first 140 bytes of the file will be sent.", flush=True,
                        #       file=sys.stderr)
                        try:
                            base = f.read().decode("utf-8")
                            ar, ln = _get_array(base)
                        except UnicodeDecodeError as e:
                            raise GgUnicodeError("Cannot send binary data from file, please use "
                                                 "the --file-transfer option.") from e
            else:
                try:
                    if msg is not None:
                        base = msg
                    elif self._script:
                        # print("Only the first 140 bytes will be sent.", flush=True,
                        # file=sys.stderr)
                        base = sys.stdin.buffer.read().decode("utf-8")
                    else:
                        raise GgArgumentsError("Wrong set of arguments.")
                    ar, ln = _get_array(base)
                    size = len(base)
                except UnicodeDecodeError as e:
                    raise GgUnicodeError("Cannot send binary data read from pipes or STDIN.") from e

            # waveform = ggwave.encode("VOX", protocolId=protocol, volume=60)
            # stream.write(waveform, len(waveform) // 4)

            crc_size = 8 if self.file_transfer_mode else 0
            if self._script:
                print("Sending data, length:", len(base) + (crc_size * ln), flush=True,
                      file=sys.stderr)
            q = 1
            totsize = 0
            if self._script:
                print(f"Piece {q-1}/{ln} {totsize} B", end="\r", flush=True, file=sys.stderr)
            t = time.time()
            for piece in ar:
                # print(piece)
                waveform = ggwave.encode(piece, protocolId=self.protocol, volume=60)
                stream.write(waveform, len(waveform) // 4)
                totsize += len(piece)
                if self._script:
                    print(f"Piece {q}/{ln} {totsize} B", end="\r", flush=True, file=sys.stderr)
                q += 1
            tt = time.time() - t
            stream.write(b'0' * 4 * self._sample_rate * 1)
            if self._script:
                print()
                print("Time taken to encode waveform:", tt, flush=True, file=sys.stderr)
            if self.file_transfer_mode and self._script:
                print("Speed (size of encoded payload + CRC):", len(base) / tt, "B/s", flush=True,
                      file=sys.stderr)
            if size and self._script:
                print("Speed (payload only):", size / tt, "B/s", flush=True, file=sys.stderr)
        except KeyboardInterrupt:
            return
        except GgIOError as e:
            if self._script:
                print(e.msg, flush=True, file=sys.stderr)
                return
            raise e
        except GgUnicodeError as e:
            if self._script:
                print(e.msg, flush=True, file=sys.stderr)
                return
            raise e
        except GgArgumentsError as e:
            if self._script:
                print(e.msg, flush=True, file=sys.stderr)
                return
            raise e
        finally:
            if stream is not None:
                stream.stop_stream()
                stream.close()
            if p is not None:
                p.terminate()



import argparse
import base64
import binascii
import io
import json
import sys
import time
from pathlib import Path
from typing import Optional, Any, BinaryIO, TextIO, Union
import pyaudio
import ggwave

from ._exceptions import GgIOError, GgChecksumError


class Receiver:
    def __init__(self, args: Optional[argparse.Namespace] = None,
                 output_file: Optional[str] = None, file_transfer: bool = False,
                 overwrite: bool = False, tot_pieces: int = -1) -> None:
        if args is not None:
            self.outputfile = args.output
            self.file_transfer_mode = args.file_transfer
            self.overwrite = args.overwrite
            self.tot_pieces: int = args.tot_pieces
        else:
            self.outputfile = output_file
            self.file_transfer_mode = file_transfer
            self.overwrite = overwrite
            self.tot_pieces = tot_pieces

    def receive(self, getdata: bool = True) -> Optional[str]:
        p: Optional[pyaudio.PyAudio] = None
        stream: Optional[pyaudio.Stream] = None
        instance: Optional[Any] = None
        file_path: Path = Path()
        output: Union[io.BytesIO | BinaryIO | TextIO]
        is_stdout = self.outputfile is None or self.outputfile == "-"

        try:
            if not is_stdout and not getdata:
                file_path = Path(self.outputfile)
                if file_path.is_file() and not self.overwrite:
                    raise GgIOError(f"File '{file_path.absolute()}' already exists, use "
                                    f"--overwrite to overwrite it.")
                output = open(file_path, "wb", buffering=0)
            elif not getdata:
                output = sys.stdout.buffer
            else:
                output = io.BytesIO()

            p = pyaudio.PyAudio()

            stream = p.open(format=pyaudio.paFloat32, channels=1, rate=48000,
                            input=True, frames_per_buffer=1024)

            ggwave.disableLog()
            par = ggwave.getDefaultParameters()
            # par["SampleRate"] = 44100
            # par["SampleRateInp"] = 44100
            # par["SampleRateOut"] = 44100
            # par["Channels"] = 1
            # par["Frequency"] = 44100
            # par["SampleWidth"] = 8192
            # par["SampleDepth"] = 8192
            # par["SampleType"] = 2
            # par["SampleChannels"] = 1
            # par["SampleFrequency"] = 44100
            instance = ggwave.init(par)

            i = 0
            started = False
            pieces = 0
            buf = ""
            size = 0
            last_crc: str = ""
            crc_file: str = ""
            start_time: float = 0

            if not getdata:
                print('Listening ... Press Ctrl+C to stop', file=sys.stderr, flush=True)
            while True:
                data = stream.read(1024, exception_on_overflow=False)
                res = ggwave.decode(instance, data)
                if res is not None:
                    st: str = res.decode("utf-8")
                    if not started and self.file_transfer_mode and st.startswith("{"):
                        # if not getdata:
                        #     print("Got Header", file=sys.stderr, flush=True)
                        js = json.loads(st)
                        pieces = js["pieces"]
                        filename = js["filename"]
                        size = js["size"]
                        crc_file = js["crc"]
                        if not getdata:
                            print(f"Got header - Filename: {filename}, Size: {size},"
                                  f" CRC32: {crc_file}, Total pieces: {pieces}",
                                  file=sys.stderr, flush=True)
                        if not getdata:
                            print(f"Piece {i}/{pieces} 0 B", end="\r", flush=True,
                                  file=sys.stderr)
                        started = True
                        start_time = time.time()
                    elif started and self.file_transfer_mode:
                        if i < pieces:
                            if i == 0:
                                last_crc = st[0:8]
                            else:
                                crc32_r = st[0:8]
                                crc32_c = binascii.crc32(buf[-132:].encode())
                                fixed_length_hex = f'{crc32_c:08x}'
                                if not fixed_length_hex == crc32_r:
                                    raise GgChecksumError(f"Received block's checksum ({fixed_length_hex}) is different from the expected: {crc32_r}.")
                            buf += st[8:]
                            i += 1
                            if not getdata:
                                print(f"Piece {i}/{pieces} {len(buf) + 8*i} B", end="\r",
                                      flush=True, file=sys.stderr)
                        else:
                            break
                    elif not self.file_transfer_mode:
                        # if not getdata:
                        #     print("Got message", file=sys.stderr, flush=True)
                        output.write(res)
                        output.flush()
                        i += 1
                        if getdata:
                            break
                        if i >= self.tot_pieces != -1:
                            break

                if i >= pieces and started:
                    last_block_len = len(buf) % 132
                    crc32_c = binascii.crc32(buf[-last_block_len:].encode())
                    fixed_length_hex= f'{crc32_c:08x}'
                    if not fixed_length_hex == last_crc:
                        raise GgChecksumError(f"Received block's checksum ({fixed_length_hex}) is different from the expected: {last_crc}.")
                    decoded_data = base64.urlsafe_b64decode(buf)
                    crc32_c = binascii.crc32(decoded_data)
                    fixed_length_hex = f'{crc32_c:08x}'
                    if not fixed_length_hex == crc_file:
                        raise GgChecksumError(f"File's checksum ({fixed_length_hex}) is different from the expected: {crc_file}.")
                    output.write(decoded_data)
                    output.flush()
                    if not getdata and self.file_transfer_mode:
                        elapsed_time = time.time() - start_time
                        print("Speed (size of encoded payload + CRC):", len(buf) / elapsed_time, "B/s", flush=True,
                              file=sys.stderr)
                        if size:
                            print("Speed (payload only):", size / elapsed_time, "B/s", flush=True, file=sys.stderr)
                    if not is_stdout and not getdata:
                        output.close()
                        stats = file_path.stat()
                        if stats.st_size != size:
                            raise GgIOError(f"File size mismatch! ({stats.st_size}, {size})")
                        if not getdata:
                            print("\nFile received, CRC correct!", file=sys.stderr, flush=True)
                    break
        except KeyboardInterrupt:
            return None
        except GgChecksumError as e:
            print(e.msg, file=sys.stderr, flush=True)
            return None
        except GgIOError as e:
            print(e.msg, file=sys.stderr, flush=True)
            return None
        finally:
            if instance is not None:
                ggwave.free(instance)
            if stream is not None:
                stream.stop_stream()
                stream.close()
            if p is not None:
                p.terminate()
        if getdata and isinstance(output, io.BytesIO):
            ret: str = output.getvalue().decode("utf-8")
            return ret
        return None



import unittest
import ggtransfer


class MyTestCase(unittest.TestCase):

    @unittest.skip("skipping test_send...")
    def test_send(self) -> None:
        s = ggtransfer.Sender(protocol=2)
        s.send("1234567890" * 15)
        # s.send()
        self.assertEqual(True, True)  # add assertion here

    @unittest.skip("skipping test_wrong_args...")
    def test_wrong_args(self) -> None:
        # noinspection PyTypeChecker
        s = ggtransfer.Sender(args="ciao", protocol=2)
        s.send("Ciao!" * 40)
        self.assertEqual(True, True)  # add assertion here

    @unittest.skip("skipping test_receive...")
    def test_receive(self) -> None:
        r = ggtransfer.Receiver(file_transfer=False)
        rr = r.receive()
        self.assertIsInstance(rr, str)
        print("-" * 30)
        print(rr)
        print("-" * 30)


if __name__ == '__main__':
    unittest.main(defaultTest="MyTestCase")
